/*
 * KS0073 SERIAL LCD DISPLAY DRIVER FOR ETHERNUT
 *
 * Copyright (C) 2004 by Frank Wallenwein.
 *
 * THIS SOFTWARE IS PROVIDED BY FRANK WALLENWEIN AND IS BASED ON 
 * ON SOFTWARE FROM egnite Software GmbH
 * IT IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright (C) 2001-2003 by egnite Software GmbH. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 */

/*
 * $Log:$
 */

#include <stdlib.h>
#include <string.h>

#include <sys/config.h>
#include <dev/ks0073.h>
#include <dev/term.h>
#include <sys/timer.h>

static u_char long_delay  = LCD_LONG_DELAY;
static u_char short_delay = LCD_SHORT_DELAY;

/*!
 * \addtogroup xgDisplay
 */
/*@{*/


/*!
 * \brief Send Command/Data to serial LCD */
static void LcdSendSerialByte(u_char theChar, u_char command)
{
unsigned char i;

	outp(inp(LCD_SID_DDR)  | (1<<LCD_SID_BIT) | (1<<LCD_SCLK_BIT) , LCD_SID_DDR);
        for ( i = 0; i<5; i++)  /* 5 Start Bits */
	{
	   sbi(LCD_SID_PORT,  LCD_SID_BIT);
	   cbi(LCD_SCLK_PORT, LCD_SCLK_BIT);
	   _NOP(); _NOP(); _NOP();  _NOP(); _NOP(); _NOP(); 
	   sbi(LCD_SCLK_PORT, LCD_SCLK_BIT);
	}
	cbi(LCD_SID_PORT,  LCD_SID_BIT); /* Clear Read/Write Bit for Write Operation  */
	cbi(LCD_SCLK_PORT, LCD_SCLK_BIT);
	_NOP(); _NOP(); _NOP();  _NOP(); _NOP(); _NOP(); 
	sbi(LCD_SCLK_PORT, LCD_SCLK_BIT);

        if ( command == 0 )
	  sbi(LCD_SID_PORT,  LCD_SID_BIT); /* Set Data/Command Bit for Data Operation     */
	else
          cbi(LCD_SID_PORT,  LCD_SID_BIT); /* Set Data/Command Bit for Command Operation  */
	cbi(LCD_SCLK_PORT, LCD_SCLK_BIT);
	_NOP(); _NOP(); _NOP();  _NOP(); _NOP(); _NOP(); 
	sbi(LCD_SCLK_PORT, LCD_SCLK_BIT);

	cbi(LCD_SID_PORT,  LCD_SID_BIT);   /* Send 0 Bit for End of start byte             */
	cbi(LCD_SCLK_PORT, LCD_SCLK_BIT);
	_NOP(); _NOP(); _NOP();  _NOP(); _NOP(); _NOP(); 
	sbi(LCD_SCLK_PORT, LCD_SCLK_BIT);

    for ( i = 0; i<4; i++)  /* 4 Lower Bits */
    {
         if ( ((theChar >> i) & 1) == 1)
           sbi(LCD_SID_PORT,  LCD_SID_BIT);
	 else
    	   cbi(LCD_SID_PORT,  LCD_SID_BIT);
	 cbi(LCD_SCLK_PORT, LCD_SCLK_BIT);
	 _NOP(); _NOP(); _NOP();  _NOP(); _NOP(); _NOP(); 
	 sbi(LCD_SCLK_PORT, LCD_SCLK_BIT);
	}

	for ( i = 0; i<4; i++)  /* 4 Zero  Bits */
	{
	   cbi(LCD_SID_PORT,  LCD_SID_BIT);
	   cbi(LCD_SCLK_PORT, LCD_SCLK_BIT);
	   _NOP(); _NOP(); _NOP();  _NOP(); _NOP(); _NOP(); 
	   sbi(LCD_SCLK_PORT, LCD_SCLK_BIT);
	}
	
    for ( i = 4; i<8; i++)  /* 4 higher Bits */
	{
	  if ( ((theChar >> i) & 1) == 1)
    	   sbi(LCD_SID_PORT,  LCD_SID_BIT);
	  else
    	  cbi(LCD_SID_PORT,  LCD_SID_BIT);
	  cbi(LCD_SCLK_PORT, LCD_SCLK_BIT);
	  _NOP(); _NOP(); _NOP();  _NOP(); _NOP(); _NOP(); 
	  sbi(LCD_SCLK_PORT, LCD_SCLK_BIT);
	}

	for ( i = 0; i<4; i++)  /* 4 Zero  Bits */
	{
	   cbi(LCD_SID_PORT,  LCD_SID_BIT);
	   cbi(LCD_SCLK_PORT, LCD_SCLK_BIT);
	   _NOP(); _NOP(); _NOP();  _NOP(); _NOP(); _NOP(); 
	   sbi(LCD_SCLK_PORT, LCD_SCLK_BIT);
	}
}

/*!
 * \brief Write data byte to LCD controller.
 */
static void LcdWriteData(u_char ch)
{
    LcdSendSerialByte(ch,0);
    NutDelay(short_delay);
}

/*!
 * \brief Write command byte to LCD controller.
 */
static void LcdWriteCmd(u_char cmd, u_char xt)
{
    LcdSendSerialByte(cmd,1);
    if(xt)
      NutDelay(xt);
}


static void LcdSetCursor(u_char pos)
{
    LcdWriteCmd(0x80 + pos, long_delay);
}

static void LcdCursorHome(void)
{
    LcdWriteCmd(0x02, long_delay);
}

static void LcdCursorLeft(void)
{
    LcdWriteCmd(0x10, short_delay);
}

static void LcdCursorRight(void)
{
    LcdWriteCmd(0x14, short_delay);
}

static void LcdClear(void)
{
    LcdWriteCmd(0x01, long_delay);
}

static void LcdCursorMode(u_char on)
{
    LcdWriteCmd(on ? 0x0F : 0x0C, long_delay);
}

static void LcdInit(void)
{
    /*
     * Send a dummy data byte.
     */
    LcdWriteData(0);
    /*
     * Initialize.
     */
    LcdWriteCmd(0x3C, 2 * long_delay);
    NutDelay(50);
    LcdWriteCmd(0x09, 2 * long_delay); /* 4 Lines */
    NutDelay(50);
    LcdWriteCmd(0x38, 2 * long_delay);
    NutDelay(50);
    /*
     * Clear display.
     */
    LcdWriteCmd(0x07, 2 * long_delay);

    /*
     * Disable display shift and set incremental mode.
     */
    LcdWriteCmd(0x06,  2 * long_delay);

    /*
     * Enable the display and a blinking cursor.
     */
    LcdWriteCmd(0x0F,  2 * long_delay);
}

/*!
 * \brief Terminal device control block structure.
 */
TERMDCB dcb_term = {
    LcdInit,            /*!< \brief Initialize display subsystem, dss_init. */
    LcdWriteData,       /*!< \brief Write display character, dss_write. */
    LcdWriteCmd,        /*!< \brief Write display command, dss_command. */
    LcdClear,           /*!< \brief Clear display, dss_clear. */
    LcdSetCursor,       /*!< \brief Set display cursor, dss_set_cursor. */
    LcdCursorHome,      /*!< \brief Set display cursor home, dss_cursor_home. */
    LcdCursorLeft,      /*!< \brief Move display cursor left, dss_cursor_left. */
    LcdCursorRight,     /*!< \brief Move display cursor right, dss_cursor_right. */
    LcdCursorMode,      /*!< \brief Switch cursor on/off, dss_cursor_mode. */
    0,                  /*!< \brief Mode flags. */
    0,                  /*!< \brief Status flags. */
    4,                  /*!< \brief Number of rows. */
    32,                 /*!< \brief Number of columns per row. */
    20,                 /*!< \brief Number of visible columns. */
    0,                  /*!< \brief Cursor row. */
    0,                  /*!< \brief Cursor column. */
    0                   /*!< \brief Display shadow memory. */
};

/*!
 * \brief LCD device information structure.
 */
NUTDEVICE devLcdSer = {
    0,              /*!< Pointer to next device. */
    {'l', 'c', 'd', 's', 'e', 'r', 0, 0, 0},      /*!< Unique device name. */
    IFTYP_STREAM,   /*!< Type of device. */
    0,              /*!< Base address. */
    0,              /*!< First interrupt number. */
    0,              /*!< Interface control block. */
    &dcb_term,      /*!< Driver control block. */
    TermInit,       /*!< Driver initialization routine. */
    TermIOCtl,      /*!< Driver specific control function. */
    0,
    TermWrite,
    TermWrite_P,
    TermOpen,
    TermClose,
    0
};

/*@}*/
